from nltk.tokenize import sent_tokenize

from wordview import logger


class DataFrameReader:
    """Reads a dataframe column and returns sentences."""

    def __init__(self, dataframe, column_name):
        """Initializes the DataFrameReader object.

        Args:
            dataframe: A pandas dataframe.
            column_name: A string representing the column name in the dataframe where corpus resides.

        Returns:
            None
        """
        if column_name not in dataframe.columns:
            raise ValueError(f"'{column_name}' not found in the dataframe.")
        self.data = dataframe[column_name]

    def get_sentences(self):
        """Returns a generator of sentences from the dataframe column.
        An nltk.sent_tokenize is applied to each row of the dataframe column
        to extract sentences from the text at each row of the specified column i.e. `column_name`.

        Args:
            None
        Returns:
            A generator of sentences.
        """
        for text in self.data:
            try:
                sentences = sent_tokenize(text)
            except Exception as E:
                logger.warning(f"Could not sentence tokenize text: {text}")
                logger.warning(E)
                continue
            for sentence in sentences:
                yield sentence



import json
import math
from typing import Dict


class PMICalculator:
    """Calculates the Pointwise Mutual Information (PMI) of an n-gram candidate."""

    def __init__(self, ngram_count_source=None, ngram_count_file_path=None):
        """Initializes a new instance of PMICalculator class.

        Args:
            ngram_count_source: A dictionary containing ngram counts.
            ngram_count_file_path: A path to a json file containing ngram counts.

        Returns:
            None
        """
        if ngram_count_source:
            self.counts = ngram_count_source
        elif ngram_count_file_path:
            self.counts = self._load_ngram_counts(ngram_count_file_path)
        else:
            raise ValueError("Either count_source or count_file_path must be provided.")

        self.total_count = sum(self.counts.values())

    def _load_ngram_counts(self, count_file_path) -> Dict[str, int]:
        with open(count_file_path, "r") as file:
            counts = json.load(file)
        return counts

    def _prob(self, ngram) -> float:
        return self.counts.get(ngram, 0) / self.total_count

    def compute_association(self, ngram) -> float:
        words = ngram.split()

        p_denominator = 1.0
        for word in words:
            p_denominator *= self._prob(word)

        p_ngram = self._prob(ngram)

        if p_denominator == 0 or p_ngram == 0:
            return float("-inf")
        else:
            return math.log(p_ngram / p_denominator, 2)



from typing import Any, Dict, List

from sentence_transformers import SentenceTransformer
from sklearn.cluster import AgglomerativeClustering, KMeans
from sklearn.feature_extraction.text import TfidfVectorizer


class Cluster:
    """Cluster text documents using various clustering algorithms and based on different vectorization technologies."""

    def __init__(self, documents: List[str], vector_model: str = "tfidf") -> None:
        """
        Args:
            documents (List): List of documents.
            vector_model (str): Vectorization technology. Default = tfidf.
        """
        self.documents = documents
        self.vector_model = vector_model
        self.vectors = self._vectorize()
        self.clusters: Dict[str, List] = {}

    def _vectorize(self):
        """Vectorizes documents in self.documents using the model specified by self.vector_model.

        Args:
            None

        Returns:
            vectors (numpy.ndarray)
        """
        if self.vector_model == "transformer":
            vectorizer = SentenceTransformer("all-MiniLM-L6-v2")
            vectors = vectorizer.encode(self.documents)
        elif self.vector_model == "tfidf":
            vectorizer = TfidfVectorizer(stop_words="english")
            vectors = vectorizer.fit_transform(self.documents)
            # cipy.sparse._csr.csr_matrix --> numpy.ndarray
            vectors = vectors.toarray()
        else:
            raise ValueError(f"Vector model {self.vector_model} is not supported.")
        return vectors

    def cluster(
        self,
        clustering_algorithm: str = "kmeans",
        n_clusters: int = 3,
        distance_threshold: Any = None,
    ) -> None:
        """Cluster documents using the algorithm specified by clustering_algorithm.

        Args:
            clustering_algorithm (str): Clustering algorithm. Default = kmeans.
                Supported algorithms: [Kmeans, AgglomerativeClustering].
            n_clusters (str): Number of clusters. Default = 3.
            distance_threshold (float): Distance threshold for AgglomerativeClustering. Default = None.

            Returns:
                None
        """
        if clustering_algorithm == "AgglomerativeClustering":
            clust_model = AgglomerativeClustering(
                n_clusters=n_clusters, distance_threshold=distance_threshold
            )
            clust_model.fit(self.vectors)
            clust_labels = clust_model.labels_
        elif clustering_algorithm == "kmeans":
            clust_model = KMeans(n_clusters=n_clusters, max_iter=100, n_init=1)
            clust_model.fit(self.vectors)
            clust_labels = clust_model.labels_
        else:
            raise ValueError(
                f"Clustering algorithm {clustering_algorithm} is not supported."
            )
        clusters: Dict[str, List] = {j: [] for j in clust_labels}
        for i in range(len(self.documents)):
            clusters[clust_labels[i]].append(self.documents[i])
        self.clusters = clusters



from typing import Dict, Iterable, Set

import numpy as np
import numpy.typing as npt
import pandas as pd
import plotly.figure_factory as ff
from scipy.stats import norm, shapiro, zscore

from wordview import logger
from wordview.anomaly import gaussianize


class NormalDistAnomalies(object):
    def __init__(
        self,
        items: Dict,
        val_name: str = "representative_value",
        gaussianization_strategy: str = "brute",
    ):
        """Identify anomalies on a normal distribution.

        Args:
            items: A dictionary of items and their representative value, such as word_count, idf, etc.
            val_name: Name of the value in the above dictionary. i.e. word_count, idf, etc. Defaults to `representative_value`.
            gaussianization_strategy: Strategy for gaussianization. Can be any of lambert, brute, or boxcox. Defaults = `brute`.

        Returns:
            None
        """
        self.val_name = val_name
        self.gaussianization_strategy = gaussianization_strategy
        self.item_value_df = pd.DataFrame(
            items.items(), columns=["item", self.val_name]
        )
        # Gaussianize values
        g_ersults = self._gaussianize_values(
            self.item_value_df[self.val_name], strategy=self.gaussianization_strategy
        )
        if shapiro(g_ersults).pvalue > 0.05:
            self.item_value_df["guassian_values"] = g_ersults
            # Calculate normal prob of gaussianized values
            dist = norm(
                loc=np.mean(self.item_value_df["guassian_values"]),
                scale=np.std(self.item_value_df["guassian_values"]),
            )
            self.item_value_df["normal_prob"] = self.item_value_df[
                "guassian_values"
            ].apply(lambda x: dist.pdf(x))
        else:
            logger.error(
                "The provided values cannot be gaussianized. Please consider using another anomaly detection method."
            )

    def anomalous_items(
        self,
        manual: bool = False,
        z: int = 3,
        prob: float = 0.001,
    ) -> Set[str]:
        """Identify anomalous items in `self.items`.

        Args:
            manual: Whether or not select redundant words using manual thresholds.
                    When set to True, manual_thresholds should be specified.
            z: Items with a z-score above this value are considered anomalous. Used only when manual is False.  Default = 3.
            prob: Probability threshold below which items are considered anomalous.

        Returns:
            An alphabetically sorted set of anomalous items.
        """
        if manual:
            anomalous_set = self._anomalous_items_manual(prob=prob)
        else:
            anomalous_set = self._anomalous_items_zscore(z_value=z)
        return set(sorted(anomalous_set))

    def _anomalous_items_manual(self, prob: float) -> Set[str]:
        """Identifies anomalous with respect to the input argument `prob` that specifies
        the probability threshold below which the items are considered anomalous.

        Args:
            prob: Probability threshold below which items are considered anomalous.

        Returns:
            Set of anomalous items.

        """
        anomalous_items = set(
            self.item_value_df[self.item_value_df["normal_prob"] < prob][
                "item"
            ].to_list()
        )
        return anomalous_items

    def _anomalous_items_zscore(self, z_value: int) -> Set[str]:
        """Identify anomalous items by looking at 'z_value'.

        Args:
            z_value: Items that are away this many standard deviation
                    from mean are considered anomalous.

        Returns:
            Set of anomalous items.
        """
        z = zscore(self.item_value_df["guassian_values"])
        self.item_value_df["zscore"] = z
        anomalies_set = set()
        for i in range(len(self.item_value_df)):
            if (
                self.item_value_df.iloc[i]["zscore"] <= -z_value
                or self.item_value_df.iloc[i]["zscore"] >= z_value
            ):
                anomalies_set.add(self.item_value_df.iloc[i]["item"])
        return anomalies_set

    def _gaussianize_values(
        self, values: Iterable[float], strategy: str
    ) -> npt.NDArray:
        """Gaussianize input values using the brute strategy.

        Args:
            values: Iterable containing numerical values.
            strategy: Strategy for gaussianization. Can be any of lambert, brute, or boxcox.

        Returns:
            numpy.NDArray containing the gaussianized distribution of `values`.
        """
        g = gaussianize.Gaussianize(strategy=strategy)
        g.fit(values)
        res = g.transform(values)
        return res



import string
import time
from statistics import median
from typing import Any, Dict, List, Tuple

import nltk
import numpy as np
import pandas
import pandas as pd
import plotly
import plotly.graph_objs as go
from langdetect import detect
from nltk.corpus import stopwords
from plotly.subplots import make_subplots
from tqdm import tqdm
from wordcloud import WordCloud, get_single_color_func

from wordview import logger


def plotly_wordcloud(token_count_dic: dict, **kwargs) -> plotly.graph_objects.Scattergl:
    """Create a world cloud trace for plotly.

    Args:
        token_count_dic (Dict): Dictionary of token to its count
        **kwargs:

    Returns:
        trace (plotly.graph_objects.Scatter)
    """
    wc_settings: Dict = kwargs.get(
        "wc_settings", {"color": "deepskyblue", "max_words": 100}
    )
    wc = WordCloud(
        color_func=get_single_color_func(wc_settings["color"]),
        max_words=wc_settings["max_words"],
    )
    wc.generate_from_frequencies(token_count_dic)
    word_list = []
    rel_freq_list = []
    freq_list = []
    fontsize_list = []
    position_list = []
    orientation_list = []
    color_list = []
    for (word, rel_freq), fontsize, position, orientation, color in wc.layout_:
        word_list.append(word)
        rel_freq_list.append(rel_freq)
        freq_list.append(token_count_dic[word])
        fontsize_list.append(fontsize)
        position_list.append(position)
        orientation_list.append(orientation)
        color_list.append(color)
    # get the positions
    x = []
    y = []
    for i in position_list:
        x.append(i[0])
        y.append(i[1])
    # get the relative occurrence frequencies
    new_freq_list = []
    for i in rel_freq_list:
        i_tmp = round(i * 100, 4)
        i_tmp = (
            i_tmp if i_tmp > 1 else 1
        )  # Plotly textfont.size in go.Scatter throws exception for values below 1.
        new_freq_list.append(i_tmp)
    try:
        trace = go.Scattergl(
            x=x,
            y=y,
            textfont=dict(size=new_freq_list, color=color_list),
            hoverinfo="text",
            hovertext=["{0}: {1}".format(w, f) for w, f in zip(word_list, freq_list)],
            mode="text",
            text=word_list,
        )
        return trace
    except Exception as E:
        logger.error(
            f"While creating the word cloud, plotly.go returned the following error \
         \n{E}\nfor relative frequencies: {rel_freq_list}\nthat were mapped to {new_freq_list}"
        )


def generate_label_plots(
    df: pandas.DataFrame, label_cols: List[Tuple], **kwargs
) -> plotly.graph_objects.Figure:
    """Generate histogram and bar plots for the labels in label_cols.

    Args:
        df (Pandas DataFrame): DataFrame that contains labels specified in label_cols.
        label_cols (list): list of tuples in the form of [('label_1', 'categorical/numerical'),
                           ('label_2', 'categorical/numerical'), ...]
        **kwargs: Additional arguments to pass to plotly.Figure.update_layout(). For more details
                  see https://plotly.com/python-api-reference/generated/plotly.graph_objects.Figure.html#plotly.graph_objects.Figure.update_layout

    Returns:
        plotly.graph_objects.Figure
    """

    if len(label_cols) == 1:
        # with titles
        # figure = make_subplots(rows=1, cols=1,subplot_titles=("Plot 1"))
        # w/o titles
        figure = make_subplots(rows=1, cols=1)
        lab_trace1 = label_plot(
            df, label_col=label_cols[0][0], label_type=label_cols[0][1]
        )
        figure.append_trace(lab_trace1, 1, 1)
        figure.update_yaxes(title_text="Count", row=2, col=2)
    elif len(label_cols) == 2:
        # with titles
        # figure = make_subplots(rows=1, cols=2,subplot_titles=("Plot 1", "Plot 2"))
        # w/o titles
        figure = make_subplots(rows=1, cols=2)
        lab_trace1 = label_plot(
            df, label_col=label_cols[0][0], label_type=label_cols[0][1]
        )
        lab_trace2 = label_plot(
            df, label_col=label_cols[1][0], label_type=label_cols[1][1]
        )
        figure.append_trace(lab_trace1, 1, 1)
        figure.append_trace(lab_trace2, 1, 2)
    elif len(label_cols) == 3:
        figure = make_subplots(rows=1, cols=3)
        lab_trace1 = label_plot(
            df, label_col=label_cols[0][0], label_type=label_cols[0][1]
        )
        lab_trace2 = label_plot(
            df, label_col=label_cols[1][0], label_type=label_cols[1][1]
        )
        lab_trace3 = label_plot(
            df, label_col=label_cols[2][0], label_type=label_cols[2][1]
        )
        figure.append_trace(lab_trace1, 1, 1)

        figure.append_trace(lab_trace2, 1, 2)
        figure.append_trace(lab_trace3, 1, 3)
    elif len(label_cols) == 4:
        figure = make_subplots(rows=2, cols=2)
        lab_trace1 = label_plot(
            df, label_col=label_cols[0][0], label_type=label_cols[0][1]
        )
        lab_trace2 = label_plot(
            df, label_col=label_cols[1][0], label_type=label_cols[1][1]
        )
        lab_trace3 = label_plot(
            df, label_col=label_cols[2][0], label_type=label_cols[2][1]
        )
        lab_trace4 = label_plot(
            df, label_col=label_cols[3][0], label_type=label_cols[3][1]
        )
        figure.append_trace(lab_trace1, 1, 1)
        figure.append_trace(lab_trace2, 1, 2)
        figure.append_trace(lab_trace3, 2, 1)
        figure.append_trace(lab_trace4, 2, 2)
    figure.update_layout(**kwargs)
    return figure


def label_plot(
    df: pandas.DataFrame, label_col: str, label_type: str
) -> plotly.graph_objects.Histogram:
    """Create a plot for label_col in df, wrt to label_type.

    Args:
        df (Pandas DataFrame): DataFrame that contains label_col.
        label_col (str): Name of the label column in df that must be plotted.
        label_type (str): Represents the type of label and consequently specifies the type of plot.
                             It can be "numerical" or "categorical".

    Returns:
        trace (plotly.graph_objects.Histogram)
    """
    if label_type == "categorical":
        values = df[label_col].unique().tolist()  # E.g. ['pos', 'neg', 'neutral']
        counts = df[label_col].value_counts()  # E.g. 1212323
        x = []
        y = []
        for v in values:
            x.append(v)
            y.append(counts[v])
        trace = go.Bar(
            x=x,
            y=y,
            name=label_col,
            marker=dict(
                color=y, coloraxis="coloraxis", line=dict(width=0.5, color="white")
            ),
        )
    elif label_type == "numerical":
        trace = go.Histogram(
            x=df[label_col],
            name=label_col,
            marker=dict(
                color=df[label_col],
                coloraxis="coloraxis",
                line=dict(width=0.5, color="white"),
            ),
        )
    else:
        raise ValueError(
            'label_col input argument must be set to either "categorical" or "numerical".'
        )
    return trace


def do_txt_analysis(
    df: pandas.DataFrame,
    text_col: str,
    language: str = "english",
    skip_stopwords_punc: bool = True,
) -> Any:
    """Generate analysis report and eitherr renders the report via Plotly show api or saves it offline to html.

    Args:
        df (pandas.DataFrame): DataFrame that contains text and labels.
        text_col (str): Name of the column that contains a tokenized text content.
        language (str): Language of the text in df[text_col]
        skip_stopwords_punc (bool): Whether or not skip stopwords and punctuations in the analysis. Default: True

    Returns:
        TxtAnalysisFields
    """

    global ftmodel

    def update_count(items_dic: dict, items: List[str]) -> None:
        """Update the corresponding count for each key in  items_dic. w.r.t. terms in items.

        Args:
            items_dic (dict): Dictionary mapping keys to their count
            items (list): List of tokens

        Returns:
            None
        """
        for t in items:
            if t in items_dic:
                items_dic[t] += 1
            else:
                items_dic[t] = 1

    def get_pos(tagged_tokens: List[Tuple[str, str]], goal_pos: str) -> List:
        """Extracts goal_pos POS tags from tagged_tokens.

        Args:
            tagged_tokens (List[Tuple(str, str)]): Contains terms and ther pos tags. E.g.
                                                   [('cat', 'NN'), ('sat', 'VBD'), ('on', 'IN'), ('mat', 'NN')]
            goal_pos (str): Pos tag to look for in tagged_tokens

        Returns:
            res (List(str)): List of tokens with goal_pos pos tag
        """
        res = []
        for pt in tagged_tokens:
            if pt[1].startswith(goal_pos):
                res.append(pt[0])
        return res

    stop_words = set(stopwords.words(language))
    punctuations = set(string.punctuation)

    doc_lengths = []
    token_to_count_dict: Dict[str, int] = {}
    NNs: Dict[str, int] = {}
    JJs: Dict[str, int] = {}
    Vs: Dict[str, int] = {}
    languages = set()

    logger.info("Processing text in %s column of the input DataFrame..." % text_col)
    for text in tqdm(df[text_col]):
        ls = detect(text).upper()
        languages.update([ls])
        try:
            tokens = text.lower().split(" ")
            doc_lengths.append(len(tokens))
            if skip_stopwords_punc:
                tokens = [
                    t for t in tokens if t not in stop_words and t not in punctuations
                ]
                update_count(token_to_count_dict, tokens)

        except Exception as e:
            logger.warning(
                "Processing entry --- %s --- lead to exception: %s" % (text, e.args[0])
            )
            continue

        postag_tokens = nltk.pos_tag(tokens)
        nouns = get_pos(postag_tokens, "NN")
        update_count(NNs, nouns)
        verbs = get_pos(postag_tokens, "VB")
        update_count(Vs, verbs)
        adjectives = get_pos(postag_tokens, "JJ")
        update_count(JJs, adjectives)

    freq_df = pd.DataFrame(
        {"tokens": token_to_count_dict.keys(), "count": token_to_count_dict.values()}
    )
    freq_df["proportion"] = freq_df["count"] / freq_df["count"].sum()
    vocab_size = freq_df.shape[0]
    n_tokens = freq_df["count"].sum()
    s = 1
    denom = np.sum(1 / (np.arange(1, n_tokens + 1) ** s))

    def classic_zipf(k, s=s, denom=denom):
        num = 1 / k**s
        res = num / denom
        return res

    logger.info("Calculating Empirical and Theoretical Zipf values...")
    freq_df["position"] = freq_df.index + 1
    start = time.time()
    freq_df["predicted_proportion"] = freq_df["position"].apply(classic_zipf)
    end = time.time()
    logger.info(
        f"Time to measure predicted proportion for {freq_df.shape[0]} rows: {end - start}"
    )

    x = np.log(freq_df["position"].values)
    y_emperical = np.log(freq_df["count"])
    y_theoritical = np.log(freq_df["predicted_proportion"] * n_tokens)

    return TxtAnalysisFields(
        doc_lengths=doc_lengths,
        zipf_x=x,
        zipf_y_emp=y_emperical,
        zipf_y_theory=y_theoritical,
        languages=languages,
        type_count=vocab_size,
        token_count=n_tokens,
        doc_count=len(doc_lengths),
        median_doc_len=median(doc_lengths),
        nns=NNs,
        jjs=JJs,
        vs=Vs,
        token_to_count_dict=token_to_count_dict,
    )


class TxtAnalysisFields:
    def __init__(
        self,
        doc_lengths,
        zipf_x,
        zipf_y_emp,
        zipf_y_theory,
        languages,
        type_count,
        token_count,
        doc_count,
        median_doc_len,
        nns,
        jjs,
        vs,
        token_to_count_dict,
    ):
        self.doc_lengths = doc_lengths
        self.zipf_x = zipf_x
        self.zipf_y_emp = zipf_y_emp
        self.zipf_y_theory = zipf_y_theory
        self.languages = languages
        self.type_count = type_count
        self.token_count = token_count
        self.doc_count = doc_count
        self.median_doc_len = median_doc_len
        self.nns = nns
        self.jjs = jjs
        self.vs = vs
        self.token_to_count_dict = token_to_count_dict



import re
import string
from re import Match
from typing import Dict, Optional

import nltk
import pandas
from nltk import RegexpParser, word_tokenize
from tabulate import tabulate  # type: ignore
from tqdm import tqdm

from wordview import logger
from wordview.io.dataframe_reader import DataFrameReader
from wordview.mwes.association_measures import PMICalculator
from wordview.mwes.patterns import DeMWEPatterns, EnMWEPatterns


# TODO use this function to ensure that the tokens are alphanumeric
def is_alphanumeric_latinscript_multigram(word: str) -> Optional[Match[str]]:
    match: Optional[Match] = re.match("[a-zA-Z0-9]{2,}", word)
    return match


class MWEPatternAssociation:
    """Extract MWE candidates from a list of tokens based on a given pattern."""

    def __init__(self, association_measure: PMICalculator, pattern: str) -> None:
        """Initializes a new instance of MWEExtractor class.

        Args:
            association_measure: An instance of an association measure class.
            pattern: A string pattern to match against the tokens. The pattern must be a string of the following form.

        Examples of user-defined patterns:
        - NP: {<DT>?<JJ>*<NN>} # Noun phrase
        - VP: {<MD>?<VB.*><NP|PP|CLAUSE>+$} # Verb phrase
        - PP: {<IN><NP>} # Prepositional phrase

        You can use multiple and/or nested patterns, separated by a newline character:
        pattern = '''
        NP: {<DT>?<JJ>*<NN>} # Noun phrase
        PROPN: {<NNP>+} # Proper noun
        ADJP: {<RB|RBR|RBS>*<JJ>} # Adjective phrase
        ADVP: {<RB.*>+<VB.*><RB.*>*} # Adverb phrase
        '''

        In this case, patterns of a clause are executed in order.  An earlier
        pattern may introduce a chunk boundary that prevents a later pattern from executing.
        """
        self.pattern = pattern
        self.association_measure = association_measure

    def _extract_mwe_candidates(self, tokens: list[str]) -> dict:
        """
        Extract variable-length MWE from tokenized input, using a user-defined POS regex pattern.

        Args:
            tokens (list[str]): A list of tokens from which mwe candidates are to be extracted.

        Returns:
            match_counter (dict[str, dict[str, int]]): A counter dictionary with count of matched strings, grouped by pattern label.
                                                    An empty list if none were found.
        """

        def get_pos_tags(tokens: list[str]) -> list[tuple[str, str]]:
            pos_tags = nltk.pos_tag(tokens)
            return pos_tags

        def validate_input() -> None:
            """Validate input argument `tokens`."""
            if not isinstance(tokens, list):
                raise TypeError(
                    f'Input argument "tokens" must be a list of string. Currently it is of type {type(self.tokens)} \
                    with a value of: {self.tokens}.'
                )
            if len(tokens) == 0:
                raise ValueError(
                    'Input argument "tokens" must be a non-empty list of string.'
                )
            if not isinstance(self.pattern, str):
                raise TypeError(
                    f'Input argument "pattern" must be a string. Currently it is of type {type(self.pattern)} \
                    with a value of: {self.pattern}.'
                )
            if len(self.pattern) == 0:
                raise ValueError(
                    'Input argument "pattern" must be a non-zero length string.'
                )

        validate_input()
        tagged_tokens: list[tuple[str, str]] = get_pos_tags(tokens)
        parser = RegexpParser(self.pattern)
        parsed_tokens = parser.parse(tagged_tokens)

        labels: list[str] = [
            rule.split(":")[0].strip() for rule in self.pattern.split("\n") if rule
        ]

        matches: dict[str, set] = {label: set() for label in labels}

        for subtree in parsed_tokens.subtrees():
            label = subtree.label()
            if label in matches:
                matches[label].add(" ".join(word for (word, tag) in subtree.leaves()))
        return matches

    def measure_candidate_association(self, tokens: list[str], threshold: float = 1.0):
        mwes: dict[str, dict[str, float]] = {}
        for mwe_type, candidate_set in self._extract_mwe_candidates(
            tokens=tokens
        ).items():
            if mwe_type not in mwes:
                mwes[mwe_type] = {}
            for mwe_candidate in candidate_set:
                association = self.association_measure.compute_association(
                    mwe_candidate
                )
                if association > threshold:
                    mwes[mwe_type][mwe_candidate] = association
        return mwes


class MWE:
    def __init__(
        self,
        df: pandas.DataFrame,
        text_column: str,
        ngram_count_source=None,
        ngram_count_file_path=None,
        language: str = "EN",
        custom_patterns: Optional[str] = None,
        only_custom_patterns: bool = False,
    ) -> None:
        """Initializes a new instance of MWE class.

        Args:
            corpus: A pandas DataFrame containing the corpus.
            text_column: The name of the column containing the text  (corpus).
            ngram_count_source: A dictionary containing ngram counts.
            ngram_count_file_path: A path to a json file containing ngram counts.
            language: The language of the corpus. Currently only 'EN' and 'DE' are supported.
            custom_pattern: pattern: A string pattern to match against the tokens. The pattern must be a string of the following form.
                Examples of user-defined patterns:
                NP: {<DT>?<JJ>*<NN>} # Noun phrase
                You can use multiple and/or nested patterns, separated by a newline character e.g.:
                custom_pattern = '''
                NP: {<DT>?<JJ>*<NN>} # Noun phrase
                VP: {<MD>?<VB.*><NP|PP|CLAUSE>+$} # Verb phrase
                PROPN: {<NNP>+} # Proper noun
                ADJP: {<RB|RBR|RBS>*<JJ>} # Adjective phrase
                ADVP: {<RB.*>+<VB.*><RB.*>*} # Adverb phrase'''
            only_custom_pattern: If True, only the custom pattern will be used to extract MWEs, otherwise, the default patterns will be used as well.

            Returns:
                None
        """
        self.language = language.upper()
        self.mwes: Dict[str, Dict[str, float]] = {}
        self.reader = DataFrameReader(df, text_column)
        self.mwe_extractor = None

        mwe_patterns: str = ""
        if language == "EN":
            for _, value in EnMWEPatterns().patterns.items():
                for v in value:
                    mwe_patterns += v + "\n"
        elif language == "DE":
            for _, value in DeMWEPatterns().patterns.items():
                for v in value:
                    mwe_patterns += v + "\n"
        else:
            raise ValueError(
                "Language not supported. Use 'EN' for English or 'DE' for German."
            )

        if custom_patterns:
            if only_custom_patterns:
                if not isinstance(only_custom_patterns, bool):
                    raise TypeError(
                        f"only_custom_patterns argument must be a boolean. Currently it is of type {type(only_custom_patterns)} \
                        with a value of: {custom_patterns}."
                    )
                mwe_patterns = custom_patterns
            else:
                mwe_patterns += "\n" + custom_patterns

        # Create an MWEPatternAssociation extractor object
        self.mwe_extractor = MWEPatternAssociation(
            association_measure=PMICalculator(
                ngram_count_source=ngram_count_source,
                ngram_count_file_path=ngram_count_file_path,
            ),
            pattern=mwe_patterns,
        )

    def extract_mwes(
        self,
        sort: bool = True,
        top_n: Optional[int] = None,
    ) -> dict[str, dict[str, float]]:
        for sentence in tqdm(self.reader.get_sentences()):
            try:
                tokens = [
                    word
                    for word in word_tokenize(sentence)
                    if word not in string.punctuation
                ]
            except Exception as E:
                logger.warning(
                    f"Could not word tokenize sentence: {sentence}.\
                            \n{E}.\
                            \nSkipping this sentence."
                )
                continue
            if tokens:
                returned_dict = self.mwe_extractor.measure_candidate_association(  # type: ignore
                    tokens=tokens
                )
                for key, inner_dict in returned_dict.items():
                    if key not in self.mwes:
                        self.mwes[key] = {}
                    self.mwes[key].update(inner_dict)

        if sort:
            for key, inner_dict in self.mwes.items():
                tmp = sorted(inner_dict.items(), key=lambda item: item[1], reverse=True)
                if top_n:
                    tmp = tmp[:top_n]
                self.mwes[key] = dict(tmp)

        return self.mwes

    def print_mwe_table(self):
        sub_tables = []
        for section, values in self.mwes.items():
            if values:
                formatted_values = {k: "{:.2f}".format(v) for k, v in values.items()}
                headers = [section, "Association"]
                table_data = list(formatted_values.items())
                table_str = tabulate(
                    table_data, headers=headers, tablefmt="double_outline"
                )
                sub_tables.append(table_str)
        final_table = "\n\n".join(sub_tables)
        print(final_table)



# TODO Currently does not support multiword terms (e.g. African American, Middle Eastern, Puerto Rican).
# TODO Change the intersection mechanism to add support for multiword terms.

# TODO Currently stemming is not applied to make the matches. For instance, Asian --> Asians etc.
# TODO For some of the words, keywords are added, but this has to be more systematically fixed.

gender_terms_en = {
    "male": [
        "man",
        "men",
        "male",
        "boy",
        "boys",
        "he",
        "his",
        "him",
        "himself",
        "gentleman",
        "guy",
        "dude",
        "lad",
    ],
    "female": [
        "woman",
        "women",
        "female",
        "girl",
        "girls",
        "she",
        "her",
        "hers",
        "herself",
        "lady",
        "gal",
    ],
    "transgender": ["transgender", "trans"],
    "nonbinary": ["non-binary", "nonbinary", "non-binary", "enby", "nb"],
}

gender_terms_de = {
    "male": [
        "Mann",
        "männlich",
        "Junge",
        "er",
        "sein",
        "sich (m)",
        "Herr",
        "Typ",
        "Kerl",
    ],
    "female": [
        "Frau",
        "weiblich",
        "Mädchen",
        "sie",
        "ihr",
        "sich (f)",
        "Dame",
        "Mädel",
    ],
    "transgender": ["transgender", "trans"],
    "nonbinary": ["non-binary", "nonbinary", "non-binary", "enby", "nb"],
}

racial_terms_en = {
    "white": ["white", "caucasian", "european", "anglo-saxon"],
    "black": ["black", "african", "afro", "afro-american", "negro"],
    "asian": [
        "asian",
        "asians",
        "oriental",
        "chinese",
        "japanese",
        "korean",
        "vietnamese",
    ],
    "latino": ["latino", "latinos", "hispanic", "latin", "mexican"],
    "indian": ["indian", "indians", "hindustani", "bharati"],
    "middle_eastern": [
        "arab",
        "arabs",
        "persian",
        "iranian",
        "iranians",
        "persians",
        "saudi",
        "emirati",
        "iraqi",
    ],
}

racial_terms_de = {
    "white": ["weiß", "kaukasisch", "europäisch"],
    "black": ["schwarz", "afrikanisch"],
    "asian": ["asiatisch", "chinesisch", "japanisch", "koreanisch", "vietnamesisch"],
    "latino": ["lateinamerikanisch", "hispanisch", "mexikanisch"],
    "indian": ["indisch"],
    "middle_eastern": ["arabisch", "persisch", "iranisch", "iraner"],
}

religion_terms_en = {
    "christian": [
        "christian",
        "christians",
        "catholic",
        "protestant",
        "baptist",
        "methodist",
        "evangelical",
    ],
    "muslim": ["muslim", "muslims", "islamic", "sunnah", "shiite", "sufi"],
    "jew": ["jew", "jews", "jewish", "hebrew", "yiddish", "zionist"],
    "hindu": ["hindu", "hinduism"],
    "buddhist": ["buddhist", "buddha", "dharma"],
    "atheist": ["atheist", "atheists", "agnostic", "non-believer", "secular"],
}

religion_terms_de = {
    "christian": ["Christ", "katholisch", "protestantisch", "evangelisch"],
    "muslim": ["Muslim", "Muslims", "islamisch", "Sunni", "Shiite"],
    "jew": ["Jude", "jüdisch", "Hebräer"],
    "hindu": ["Hindu"],
    "buddhist": ["Buddhist", "Buddha"],
    "atheist": ["Atheist", "Agnostiker", "säkular"],
}


def get_terms(language, category):
    """
    Fetches the set of terms for a given language and category.

    Parameters:
    - language (str): 'en' or 'de'
    - category (str): 'gender', 'racial', or 'religion'

    Returns:
    - dict: Dictionary of terms for the given language and category
    """
    if language == "en":
        if category == "gender":
            return gender_terms_en
        elif category == "racial":
            return racial_terms_en
        elif category == "religion":
            return religion_terms_en
    elif language == "de":
        if category == "gender":
            return gender_terms_de
        elif category == "racial":
            return racial_terms_de
        elif category == "religion":
            return religion_terms_de
    else:
        raise ValueError(f"Unsupported language: {language}")



import string

import plotly.graph_objects as go
from nltk import word_tokenize
from plotly.subplots import make_subplots
from sentence_transformers import SentenceTransformer
from tabulate import tabulate  # type: ignore
from tqdm import tqdm
from transformers import BertForSequenceClassification, BertTokenizer

from wordview import logger
from wordview.bias_analysis import bias_terms
from wordview.io.dataframe_reader import DataFrameReader


class BiasDetector:
    def __init__(self, df, text_column):
        # bert-base-multilingual-uncased-sentiment supports English, Dutch, German, French, Spanish, and Italian.
        self.sentiment_model = BertForSequenceClassification.from_pretrained(
            "nlptown/bert-base-multilingual-uncased-sentiment"
        )
        self.sentiment_tokenizer = BertTokenizer.from_pretrained(
            "nlptown/bert-base-multilingual-uncased-sentiment"
        )
        self.embedding_model = SentenceTransformer(
            "distiluse-base-multilingual-cased-v2"
        )
        self.reader = DataFrameReader(df, text_column)
        self.biases = {}

    def _calculate_association(self, sentence, category_terms):
        sentiments = []
        for term in category_terms:
            inputs = self.sentiment_tokenizer.encode_plus(
                f"{term} {sentence}",
                return_tensors="pt",
                truncation=True,
                max_length=512,
            )
            output = self.sentiment_model(**inputs)
            sentiment_class = output[0].argmax(1).item()
            sentiments.append(sentiment_class)
        avg_sentiment = sum(sentiments) / len(sentiments) if sentiments else 0
        return avg_sentiment

    def _detect_bias_category(self, bias_category):
        biases = {}
        for category_type, category_terms in bias_category.items():
            category_type_avg_sentiment = 0
            n = 0
            for sentence in tqdm(self.reader.get_sentences()):
                try:
                    tokens = [
                        word.lower()
                        for word in word_tokenize(sentence)
                        # TODO support languages other than English
                        if word not in string.punctuation
                    ]
                except Exception as E:
                    logger.warning(
                        f"Could not word tokenize sentence: {sentence}.\
                                \n{E}.\
                                \nSkipping this sentence."
                    )
                    continue
                if tokens:
                    intersection = set(tokens).intersection(
                        [cat_term.lower() for cat_term in category_terms]
                    )
                    if intersection:
                        association_result = self._calculate_association(
                            category_terms=intersection, sentence=sentence
                        )
                        category_type_avg_sentiment += association_result
                        n += 1
            category_type_avg_sentiment = (
                (category_type_avg_sentiment / n) if n > 0 else "-inf"
            )
            biases[category_type] = category_type_avg_sentiment
        return biases

    def show_bias_plot(self):
        categories = list(self.biases.keys())
        fig = make_subplots(
            rows=len(categories),
            cols=1,
            subplot_titles=[cat.capitalize() for cat in categories],
        )
        colorscale = [
            [0.0, "#8B0000"],  # Very Negative
            [0.25, "#FF4500"],  # Negative
            [0.5, "yellow"],  # Neutral
            [0.75, "#ADFF2F"],  # Positive
            [1.0, "#006400"],  # Very Positive
        ]
        for index, (category, sub_data) in enumerate(self.biases.items()):
            labels = list(sub_data.keys())
            labels = [label.capitalize() for label in labels]
            values = [val if val != "-inf" else None for val in sub_data.values()]

            heatmap = go.Heatmap(
                z=[values],
                x=labels,
                zmin=0,
                zmax=4,
                colorscale=colorscale,
                showscale=True,
                colorbar_title="Bias",
                colorbar=dict(
                    tickvals=[0, 1, 2, 3, 4],
                    ticktext=[
                        "Very Negative",
                        "Negative",
                        "Neutral",
                        "Positive",
                        "Very Positive",
                    ],
                    tickfont=dict(size=16),
                    titlefont=dict(size=18),
                ),
                zauto=False,  # Prevents auto scaling
            )
            fig.add_trace(heatmap, row=index + 1, col=1)
            fig.update_yaxes(showgrid=False, showticklabels=False)
            fig.update_xaxes(showgrid=False, tickfont=dict(size=16))
        fig.update_layout(
            title="Bias Scores Across Categories",
            title_font_size=20,  # Increase title font size
            title_x=0.5,  # Center main title
            width=1000,  # Fixed width
            height=250 * len(categories),  # Adjust height based on number of categories
            plot_bgcolor="white",  # Set background color to white
        )
        # Increase font size of subplot titles.
        fig.update_annotations(font_size=18)
        fig.show()

    def print_bias_table(self):
        def sentiment_to_bias(val):
            if val == "-inf":
                return "Unknown"
            elif val == 0:
                return "Very Negative"
            elif val == 1:
                return "Negative"
            elif val == 2:
                return "Neutral"
            elif val == 3:
                return "Positive"
            elif val == 4:
                return "Very Positive"

        sub_tables = []
        for section, values in self.biases.items():
            if values:
                formatted_values = {
                    k.capitalize(): sentiment_to_bias(v) for k, v in values.items()
                }
                headers = [section.capitalize(), "Bias"]
                table_data = list(formatted_values.items())
                table_str = tabulate(
                    table_data, headers=headers, tablefmt="double_outline"
                )
                sub_tables.append(table_str)
        final_table = "\n\n".join(sub_tables)
        print(final_table)

    def detect_bias(self, language="en"):
        gender_categories = bias_terms.get_terms(language, "gender")
        racial_categories = bias_terms.get_terms(language, "racial")
        religion_categories = bias_terms.get_terms(language, "religion")

        self.biases = {
            "gender": self._detect_bias_category(bias_category=gender_categories),
            "racial": self._detect_bias_category(bias_category=racial_categories),
            "religion": self._detect_bias_category(bias_category=religion_categories),
        }
        return self.biases



import re
from typing import Dict, Set

from nltk import word_tokenize


def clean_text(
    text: str,
    keep_pattern: str = "[a-zA-Z0-9!.,?]",
    drop_patterns: Set[str] = set([]),
    replace: Dict = {},
    remove_emojis=False,
    remove_blanks=False,
    maxlen: int = 15,
    lower=False,
) -> str:
    """Tokenize and clean text, based on keep_pattern, drop_patterns,
    replace patterns, and other input arguments.

    Args:
        text (str): Input text.
        keep_pattern (str): Allowed patterns e.g. [a-zA-Z]. Defaults to "[a-zA-Z0-9!.,?]".
        drop_patterns (set): Set of patterns that should be dropped from text.
        replace (dict): Dictionary of to_be_replaced_pattern: replaced_with. E.g. {[0-9]+: NUM}
        remove_emojis (bool):‌ Whether or not to remove emojis. Defaults to False.
        remove_blanks (bool): Relevant for text scraped from Web or HTML tags. Remove excessive blank lines and whitespaces. Defaults to False.
        maxlen (int): Maximum length of a token. Defaults to 15.
        lower (bool): Whether or nor lowercase the text at the end.

    Returns:
        out_text (string): Tokenized and cleaned up text with respect to all above criteria.
    """
    if not isinstance(text, str):
        raise TypeError("Input must be a string.")
    if len(text) == 0:
        raise ValueError("Input must be a non empty string.")

    if remove_emojis:
        text = remove_emojis(text)

    if remove_blanks:
        text = re.sub(" {2,}", " ", text)
        text = re.sub("\n{2,}", "\n", text)

    # Drop unwanted tokens: Replace them with space, then replace resulting \s{2, } with one space
    for d in drop_patterns:  # d = e.g. <br/>
        if re.search(d, text):
            text = re.sub(d, " ", text)
    text = re.sub(" {2,}", " ", text)

    for k, v in replace.items():
        if re.search(k, text):
            text = re.sub(k, v, text)

    tokens = word_tokenize(text)
    out_tokens = []
    for t in tokens:
        if len(t) < maxlen:
            if re.match(keep_pattern, t):
                out_tokens.append(t)
    out_text = " ".join(out_tokens)
    if lower:
        out_text = out_text.lower()
    return out_text


def remove_emojis(text: str) -> str:
    """Remove Emojis from `text`.

    Args:
        text (str): Input text to remove emojis from.

    Returns:
        text with emojis removed.
    """
    emoj = re.compile(
        "["
        "\U0001F600-\U0001F64F"  # emoticons
        "\U0001F300-\U0001F5FF"  # symbols & pictographs
        "\U0001F680-\U0001F6FF"  # transport & map symbols
        "\U0001F1E0-\U0001F1FF"  # flags (iOS)
        "\U00002500-\U00002BEF"  # chinese char
        "\U00002702-\U000027B0"
        "\U00002702-\U000027B0"
        "\U000024C2-\U0001F251"
        "\U0001f926-\U0001f937"
        "\U00010000-\U0010ffff"
        "\u2640-\u2642"
        "\u2600-\u2B55"
        "\u200d"
        "\u23cf"
        "\u23e9"
        "\u231a"
        "\ufe0f"  # dingbats
        "\u3030"
        "]+",
        re.UNICODE,
    )
    return re.sub(emoj, "", text)



"""
Transform data so that it is approximately normally distributed

This code written by Greg Ver Steeg, 2015.
"""

import numpy as np
import sklearn
from scipy.optimize import fmin  # TODO: Explore efficacy of other opt. methods
from scipy.special import lambertw
from scipy.stats import boxcox, kurtosis, norm, rankdata

np.seterr(all="warn")


class Gaussianize(sklearn.base.TransformerMixin):
    """
    Gaussianize data using various methods.

    Conventions
    ----------
    This class is a wrapper that follows sklearn naming/style (e.g. fit(X) to train).
    In this code, x is the input, y is the output. But in the functions outside the class, I follow
    Georg's convention that Y is the input and X is the output (Gaussianized) data.

    Parameters
    ----------
    tol : float, default = 1e-4

    max_iter : int, default = 100
        Maximum number of iterations to search for correct parameters of Lambert transform.

    strategy : str, default='lambert'
        Possibilities are 'lambert'[1], 'brute'[2] and 'boxcox'[3].

    Attributes
    ----------
    coefs_ : list of tuples
        For each variable, we have transformation parameters.
        For Lambert, e.g., a tuple consisting of (mu, sigma, delta), corresponding to the parameters of the
        appropriate Lambert transform. Eq. 6 and 8 in the paper below.

    References
    ----------
    [1] Georg Goerg. The Lambert Way to Gaussianize heavy tailed data with
                        the inverse of Tukey's h transformation as a special case
        Author generously provides code in R: https://cran.r-project.org/web/packages/LambertW/
    [2] Valero Laparra, Gustavo Camps-Valls, and Jesus Malo. Iterative Gaussianization: From ICA to Random Rotations
    [3] Box cox transformation and references: https://en.wikipedia.org/wiki/Power_transform
    """

    def __init__(self, tol=1.22e-4, max_iter=100, verbose=False, strategy="lambert"):
        self.tol = tol
        self.max_iter = max_iter
        self.strategy = strategy
        self.coefs_ = []  # Store tau for each transformed variable
        self.verbose = verbose

    def fit(self, x, y=None):
        """Fit a Gaussianizing transformation to each variable/column in x."""
        x = np.asarray(x)
        if len(x.shape) == 1:
            x = x[:, np.newaxis]
        elif len(x.shape) != 2:
            print(
                "Data should be a 1-d list of samples to transform or a 2d array with samples as rows."
            )

        if self.strategy == "lambert":
            if self.verbose:
                print("Gaussianizing with Lambert method")
            for x_i in x.T:
                self.coefs_.append(igmm(x_i, tol=self.tol, max_iter=self.max_iter))
        elif self.strategy == "brute":
            for x_i in x.T:
                self.coefs_.append(
                    None
                )  # TODO: In principle, we could store parameters to do a quasi-invert
        elif self.strategy == "boxcox":
            for x_i in x.T:
                self.coefs_.append(boxcox(x_i)[1])
        else:
            raise NotImplementedError
        return self

    def transform(self, x):
        """Transform new data using a previously learned Gaussianization model."""
        x = np.asarray(x)
        if len(x.shape) == 1:
            x = x[:, np.newaxis]
        elif len(x.shape) != 2:
            print(
                "Data should be a 1-d list of samples to transform or a 2d array with samples as rows."
            )
        if x.shape[1] != len(self.coefs_):
            print(
                "%d variables in test data, but %d variables were in training data."
                % (x.shape[1], len(self.coefs_))
            )

        if self.strategy == "lambert":
            return np.array([w_t(x_i, tau_i) for x_i, tau_i in zip(x.T, self.coefs_)]).T
        elif self.strategy == "brute":
            return np.array(
                [norm.ppf((rankdata(x_i) - 0.5) / len(x_i)) for x_i in x.T]
            ).T
        elif self.strategy == "boxcox":
            return np.array(
                [boxcox(x_i, lmbda=lmbda_i) for x_i, lmbda_i in zip(x.T, self.coefs_)]
            ).T
        else:
            raise NotImplementedError

    def inverse_transform(self, y):
        """Recover original data from Gaussianized data."""
        if self.strategy == "lambert":
            return np.array(
                [inverse(y_i, tau_i) for y_i, tau_i in zip(y.T, self.coefs_)]
            ).T
        elif self.strategy == "boxcox":
            return np.array(
                [
                    (1.0 + lmbda_i * y_i) ** (1.0 / lmbda_i)
                    for y_i, lmbda_i in zip(y.T, self.coefs_)
                ]
            ).T
        else:
            print("Inversion not supported for this gaussianization transform.")
            raise NotImplementedError

    def qqplot(self, x, prefix="qq"):
        """Show qq plots compared to normal before and after the transform."""
        from matplotlib import pylab
        from scipy.stats import probplot

        y = self.transform(x)

        for i, (x_i, y_i) in enumerate(zip(x.T, y.T)):
            probplot(x_i, dist="norm", plot=pylab)
            pylab.savefig(prefix + "_%d_before.png" % i)
            pylab.clf()

            probplot(y_i, dist="norm", plot=pylab)
            pylab.savefig(prefix + "_%d_after.png" % i)
            pylab.clf()


def w_d(z, delta):
    # Eq. 9
    if delta < 1e-6:
        return z
    return np.sign(z) * np.sqrt(np.real(lambertw(delta * z**2)) / delta)


def w_t(y, tau):
    # Eq. 8
    return tau[0] + tau[1] * w_d((y - tau[0]) / tau[1], tau[2])


def inverse(x, tau):
    # Eq. 6
    u = (x - tau[0]) / tau[1]
    return tau[0] + tau[1] * (u * np.exp(u * u * (tau[2] * 0.5)))


def igmm(y, tol=1.22e-4, max_iter=100):
    # Infer mu, sigma, delta using IGMM in Alg.2, Appendix C
    if np.std(y) < 1e-4:
        return np.mean(y), np.std(y).clip(1e-4), 0
    delta0 = delta_init(y)
    tau1 = (np.median(y), np.std(y) * (1.0 - 2.0 * delta0) ** 0.75, delta0)
    for k in range(max_iter):
        tau0 = tau1
        z = (y - tau1[0]) / tau1[1]
        delta1 = delta_gmm(z)
        x = tau0[0] + tau1[1] * w_d(z, delta1)
        mu1, sigma1 = np.mean(x), np.std(x)
        tau1 = (mu1, sigma1, delta1)

        if np.linalg.norm(np.array(tau1) - np.array(tau0)) < tol:
            break
        else:
            if k == max_iter - 1:
                print(
                    "Warning: No convergence after %d iterations. Increase max_iter."
                    % max_iter
                )
    return tau1


def delta_gmm(z):
    # Alg. 1, Appendix C
    delta0 = delta_init(z)

    def func(q):
        u = w_d(z, np.exp(q))
        if not np.all(np.isfinite(u)):
            return 0.0
        else:
            k = kurtosis(u, fisher=True, bias=False) ** 2
            if not np.isfinite(k) or k > 1e10:
                return 1e10
            else:
                return k

    res = fmin(func, np.log(delta0), disp=0)
    return np.around(np.exp(res[-1]), 6)


def delta_init(z):
    gamma = kurtosis(z, fisher=False, bias=False)
    with np.errstate(all="ignore"):
        delta0 = np.clip(1.0 / 66 * (np.sqrt(66 * gamma - 162.0) - 6.0), 0.01, 0.48)
    if not np.isfinite(delta0):
        delta0 = 0.01
    return delta0



import json
import os
import string
from collections import defaultdict
from datetime import datetime
from typing import Dict, Optional

import pandas as pd
from nltk.tokenize import word_tokenize
from nltk.util import ngrams

from wordview import logger
from wordview.io.dataframe_reader import DataFrameReader


class NgramExtractor:
    """Extracts n-grams from a dataframe.

    Example:
    >>> df = pd.DataFrame({
    ...     'text': [
    ...         'This is a sample sentence. Here is another one!',
    ...         'Another sample sentence here. And yet another one.',
    ...         'Yet another sample! This continues.'
    ...     ]
    ... })
    >>> extractor = NgramExtractor(df, 'text')
    >>> extractor.extract_ngrams()
    >>> ngram_counts = extractor.get_ngram_counts()
    >>> print(ngram_counts)
    """

    def __init__(self, dataframe: pd.DataFrame, column_name: str):
        """Initializes a new instance of NgramExtractor class.

        Args:
            dataframe (pd.DataFrame): A pandas dataframe.
            column_name (str): The name of the text column where corpus resides.

        Returns:
            None
        """
        self.reader = DataFrameReader(dataframe, column_name)
        self.ngram_counts: Dict[str, int] = defaultdict(int)

    def extract_ngrams(self, n: int = 4):
        """Extracts n-grams from the text column of the dataframe.

        Args:
            n (int): The maximum number of words in an n-gram. Defaults to 4.

        Returns:
            None
        """
        for sentence in self.reader.get_sentences():
            try:
                tokens = [
                    word
                    for word in word_tokenize(sentence)
                    if word not in string.punctuation
                ]
            except Exception as E:
                logger.warning(
                    f"Could not word tokenize sentence: {sentence}.\
                               \n{E}.\
                               \nSkipping this sentence."
                )
                continue

            for i in range(1, n + 1):
                for ngram in ngrams(tokens, i):
                    # Convert the n-gram tuple to a space-separated string
                    ngram_str = " ".join(ngram)
                    self.ngram_counts[ngram_str] += 1

    def get_ngram_counts(
        self, ngram_count_file_path: Optional[str] = None, overwrite: bool = True
    ) -> Dict[str, int]:
        """Returns a dictionary of n-grams and their counts.

        Args:
            ngram_count_file_path (Optional[str]): The path to the file where n-gram counts will be saved.
                                                    If not provided, the n-gram counts will not be saved to file.
            overwrite (bool): Whether to overwrite the file if it already exists. Defaults to True.

        Returns:
            ngram_counts (Dict[str, int]): A dictionary of n-grams and their counts.
                                           E.g. {'This': 2, 'is': 2, 'a': 1, 'sample': 3}
        """
        if ngram_count_file_path:
            directory = os.path.dirname(ngram_count_file_path)
            if directory:
                os.makedirs(directory, exist_ok=True)
            file_path = ngram_count_file_path
            if os.path.exists(file_path):
                if not overwrite:
                    base, ext = os.path.splitext(file_path)
                    timestamp = datetime.now().strftime("%Y%m%d%H%M%S")
                    file_path = f"{base}_{timestamp}{ext}"

            with open(file_path, "w") as file:
                json.dump(self.ngram_counts, file, ensure_ascii=False, indent=4)

        return dict(self.ngram_counts)



from typing import Dict, List


class EnMWEPatterns:
    patterns: Dict[str, List[str]] = {}

    def __init__(self, mwe_types=["LVC", "NC2", "NC3", "ANC2", "ANC3", "VPC"]):
        if "LVC" in mwe_types:
            self.patterns["LVC"] = [
                "LVC: {<VB*><DT><\\w+>}",
            ]
        if "NC2" in mwe_types:
            self.patterns["NC2"] = [
                "NC2: {<NN|NNS><NN|NNS>}",
            ]
        if "NC3" in mwe_types:
            self.patterns["NC3"] = [
                "NC3: {<NN|NNS><NN|NNS><NN|NNS>}",
            ]
        if "ANC2" in mwe_types:
            self.patterns["ANC2"] = [
                "ANC2: {<JJ><NN|NNS>}",
            ]
        if "ANC3" in mwe_types:
            self.patterns["ANC3"] = ["ANC3: {<JJ><JJ><NN|NNS>}"]
        if "VPC" in mwe_types:
            self.patterns["VPC"] = [
                "VPC: {<VB|VBP><RP>}",
            ]


class DeMWEPatterns:
    patterns: Dict[str, List[str]] = {}

    def __init__(self, mwe_types=["LVC", "NC2", "NC3", "ANC2", "ANC3", "VPC"]):
        if "LVC" in mwe_types:
            self.patterns["LVC"] = [
                "LVC: {<VB*><DT><\\w+>}",
            ]
        # Define the patterns for 2 and 3-word noun compounds (e.g., "Hausaufgaben", "Fußballplatz")
        if "NC2" in mwe_types:
            self.patterns["NC2"] = [
                "NC2: {<NN|NNS><NN|NNS>}",
            ]
        if "NC3" in mwe_types:
            self.patterns["NC3"] = [
                "NC3: {<NN|NNS><NN|NNS><NN|NNS>}",
            ]
        if "ANC2" in mwe_types:
            self.patterns["ANC2"] = [
                "ANC2: {<JJ><NN|NNS>}",
            ]
        if "ANC3" in mwe_types:
            self.patterns["ANC3"] = ["ANC3: {<JJ><JJ><NN|NNS>}"]
        # Define the patterns for verb particle constructions (e.g., "aufstehen", "zurückkommen")
        if "VPC" in mwe_types:
            self.patterns["VPC"] = [
                "VPC: {<VB|VBP><RP>}",
            ]



from typing import Dict, List, Set, Tuple

import pandas
import plotly.figure_factory as ff
import plotly.graph_objs as go
from tabulate import tabulate  # type: ignore

from wordview.text_analysis.core import (
    do_txt_analysis,
    generate_label_plots,
    plotly_wordcloud,
)


class TextStatsPlots:
    """
    Represents Text Statistics and Plots.
    """

    def __init__(
        self,
        df: pandas.DataFrame,
        text_column: str,
        distributions: Set = {"doc_len", "word_frequency_zipf"},
        pos_tags: Set = {"NN", "VB", "JJ"},
    ) -> None:
        """Initialize a new TextStatsPlots object with the given arguments.

        Args:
            df (pandas.DataFrame): DataFrame with a text_column that contains the corpus.
            text_column (str): Specifies the column of DataFrame where text data resides.
            distributions (Set): Set of distribution types to generate and plot. Available distributions are:
                doc_len: document lengths, word_frequency_zipf: Zipfian word frequency distribution.
                Default = {"doc_len", "word_frequency_zipf"}.
            pos_tags (Set): = Set of POS tags for which world cloud is shown. Default = {"NN", "VB", "JJ"}.

        Returns:
            None
        """
        self.df = df
        self.analysis = do_txt_analysis(df=self.df, text_col=text_column)
        self.distributions = distributions
        self.pos_tags = pos_tags
        self.languages = self.analysis.languages
        self.type_count = self.analysis.type_count
        self.token_count = self.analysis.token_count
        self.num_docs = self.analysis.doc_count
        self.median_doc_len = self.analysis.median_doc_len
        self.num_nns = len(self.analysis.nns)
        self.num_jjs = len(self.analysis.jjs)
        self.num_vbs = len(self.analysis.vs)

    def _create_dist_plots(self, **kwargs) -> Dict[str, go.Figure]:
        """Create distribution plots for items in `self.distributions`.

        Args:
            **kwargs: Additional arguments to be passed to the plotly figure factory.
                      For available settings see: https://plotly.com/python/reference/layout/

        Returns:
            Dictionary of distribution names to plotly go.Figure objects for that distribution.
        """
        res = {}
        if "doc_len" in self.distributions:
            fig_doc_len_dist = ff.create_distplot(
                [self.analysis.doc_lengths], group_labels=["distplot"], colors=["blue"]
            )
            res["doc_len"] = fig_doc_len_dist

        if "word_frequency_zipf" in self.distributions:
            fig_w_freq = go.Figure()
            fig_w_freq.add_trace(
                go.Scattergl(
                    x=self.analysis.zipf_x,
                    y=self.analysis.zipf_y_emp,
                    mode="markers",
                    marker=dict(
                        color=self.analysis.zipf_x,
                        colorscale="Tealgrn",
                    ),
                )
            )
            fig_w_freq.add_trace(
                go.Scattergl(
                    x=self.analysis.zipf_x,
                    y=self.analysis.zipf_y_theory,
                    mode="markers",
                    marker=dict(color=self.analysis.zipf_x, colorscale="Reds"),
                )
            )
            res["word_frequency_zipf"] = fig_w_freq
        dist_plot_settings = kwargs.get("plot_settings", {"showlegend": False})
        for _, fig in res.items():
            fig.update_layout(dist_plot_settings)

        return res

    def show_distplot(self, distribution: str, **kwargs) -> None:
        """Shows distribution plots for `dist`.

        Args:
            dist (str): The distribution for which the plot is to be shown.
                        Can be either of: doc_len" or "word_frequency_zipf.
            **kwargs: Additional arguments to be passed to self._create_dist_plots and then plotly figure factory.
                      For available settings see: https://plotly.com/python/reference/layout/

        Returns:
            None
        """
        self._create_dist_plots(**kwargs)[distribution].show()

    def _create_pos_plots(
        self,
        pos: str,
        **kwargs,
    ) -> go.Figure:
        """Create plots for the POS tags specified in items in `self.pos_tags`.

        Args:
            pos (str): The POS tag for which the plot is to be shown.

        Returns:
            Dictionary of POS tags to plotly go.Figure objects.

        """
        word_cloud_plot_mandatory_settings = {
            "showlegend": False,
            "xaxis_showgrid": False,
            "yaxis_showgrid": False,
            "xaxis_zeroline": False,
            "yaxis_zeroline": False,
            "yaxis_visible": False,
            "yaxis_showticklabels": False,
            "xaxis_visible": False,
            "xaxis_showticklabels": False,
        }
        plot_settings = kwargs.get("plot_settings", {})
        plot_settings = {**word_cloud_plot_mandatory_settings, **plot_settings}
        if pos == "NN" and "NN" in self.pos_tags:
            return go.Figure(
                plotly_wordcloud(token_count_dic=self.analysis.nns, **kwargs)
            ).update_layout(plot_settings)
        elif pos == "JJ" and "JJ" in self.pos_tags:
            return go.Figure(
                plotly_wordcloud(token_count_dic=self.analysis.jjs, **kwargs)
            ).update_layout(plot_settings)
        elif pos == "VB" and "VB" in self.pos_tags:
            return go.Figure(
                plotly_wordcloud(token_count_dic=self.analysis.vs, **kwargs)
            ).update_layout(plot_settings)
        else:
            raise ValueError(
                f"Invalid value for pos: {pos}. Valid values are: {self.pos_tags}"
            )

    def show_word_clouds(self, pos: str, **kwargs) -> None:
        """Shows POS word clouds.

        Args:
            pos (str): Type of POS. Can be any of: [NN, JJ, VB].
            **kwargs: Keyword arguments to be passed to self._create_pos_plots() and wordview.text_analysis.core.plotly_wordcloud().
            This includes:
            - plot_settings: Dictionary of form: for self._create_pos_plots(). For available settings see: https://plotly.com/python/reference/layout/.
            - wc_settings: Dictionary of form: {"color": "<color>", "max_words": int} for core.plotly_wordcloud(). Accepted values are color strings as usable by PIL/Pillow.

        Returns:
            None
        """
        self._create_pos_plots(pos=pos, **kwargs).show()

    def show_stats(self) -> None:
        """Print dataset statistics, including:
        Language/s
        Number of unique words
        Number of all words
        Number of documents
        Median document length
        Number of nouns
        Number of adjectives
        Number of verbs.
        """
        table = tabulate(
            [
                ["Language/s", ", ".join(self.languages)],
                ["Unique Words", f"{self.type_count:,d}"],
                ["All Words", f"{self.token_count:,d}"],
                ["Documents", f"{self.num_docs:,d}"],
                ["Median Doc Length", self.median_doc_len],
                ["Nouns", f"{self.num_nns:,d}"],
                ["Adjectives", f"{self.num_jjs:,d}"],
                ["Verbs", f"{self.num_vbs:,d}"],
            ],
            tablefmt="simple_grid",
        )
        print(table)

    def show_insights(self):
        "Show topics, MWEs, clusters,"
        raise NotImplementedError


class LabelStatsPlots:
    """
    Represents Label Statistics and Plots.
    """

    def __init__(
        self,
        df: pandas.DataFrame,
        label_columns: List[Tuple],
    ) -> None:
        """Initialize a new LabelStatsPlots object with the given arguments.

        Args:
            df (pandas.DataFrame): DataFrame with one or more label column/s.
            label_columns (List): List of tuples (column_name, label_type) that specify a label column and its type (categorical or numerical).

        Returns:
            None
        """
        self.df = df
        self.label_columns = label_columns

    def show_label_plots(self, **kwargs) -> None:
        """Renders label plots for columns specified in `self.label_columns`.

        Args:
            **kwargs: Additional arguments to be passed to generate_label_plots() to be used by plotly.Figure.update_layout(). For more details
                  see https://plotly.com/python-api-reference/generated/plotly.graph_objects.Figure.html#plotly.graph_objects.Figure.update_layout

        Returns:
            None
        """
        generate_label_plots(self.df, label_cols=self.label_columns, **kwargs).show()
